/**

    Copyright 2007, 2010 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic.workflow.api.jbpm;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.atomic.AtomicBoolean;

import org.apache.pdfbox.pdmodel.fdf.FDFDocument;
import org.eclipse.ebpm.workflow.api.IControlAPI;
import org.eclipse.ebpm.workflow.api.IProcessEngine;
import org.eclipse.ebpm.workflow.api.ModifiedMessage;
import org.eclipse.ebpm.workflow.api.SpagicWorkflowException;
import org.eclipse.ebpm.workflow.api.Variable;
import org.hibernate.Session;
import org.jbpm.JbpmConfiguration;
import org.jbpm.JbpmContext;
import org.jbpm.graph.exe.ExecutionContext;
import org.jbpm.graph.exe.Token;
import org.jbpm.taskmgmt.exe.TaskInstance;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.spagic.metadb.base.ProcessInstanceStateConstants;
import org.spagic.metadb.dbutils.AuditDBUtils;
import org.spagic.metadb.model.FormData;
import org.spagic.metadb.model.LogJournal;
import org.spagic.metadb.model.PoolInstance;
import org.spagic.metadb.model.ProcessInstance;
import org.spagic.metadb.model.TransitionInstance;
import org.spagic.workflow.api.jbpm.internal.cmd.CreateProcessInstanceCommand;

/**
 * The Class ControlAPI.
 */
public class ControlAPI implements IControlAPI {

	/** The Constant log. */
	
	private static Logger log = LoggerFactory.getLogger(ControlAPI.class);

	/** The Constant ANONYMOUS_POOL_TYPE. */
	private static final String ANONYMOUS_POOL_TYPE = "ANONYMOUS".intern();
	private static final String TASK_VARIABLE_NOTATION = "_[Task]_".intern();

	private static final boolean START_TASK = false;
	private static final boolean END_TASK = true;

	/**
	 * The name of the variable that in JBPM carries the XML Messages
	 */
	public static final String XML_MESSAGE = "XML_MESSAGE";

	/** The jbpm configuration. */
	private JbpmConfiguration jbpmConfiguration = null;

	/** The jbpm engine. */
	private ProcessEngine jbpmEngine = null;

	/**
	 * Instantiates a new control api.
	 * 
	 * @param jbpmConfiguration
	 *          the jbpm configuration
	 * @param processEngine
	 *          the process engine
	 */
	public ControlAPI(JbpmConfiguration jbpmConfiguration, IProcessEngine processEngine) {
		super();
		this.jbpmConfiguration = jbpmConfiguration;
		this.jbpmEngine = (ProcessEngine) processEngine;
	}

	/**
	 * @see org.eclipse.ebpm.workflow.api.IControlAPI#abort(long)
	 */
	public void abort(long processInstanceId) {
		
		ProcessInstanceExecutionContext iec = ProcessEngine.buildProcessInstanceExecutionContext(processInstanceId);
		ProcessEngine.doWithinJBPM("ControlAPI.abort()", jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session spagicSession, ProcessInstanceExecutionContext instanceExecutionContext) {
						
						org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance = jbpmContext.getProcessInstanceForUpdate(instanceExecutionContext.getJbpmProcessInstanceId());
						jbpmProcessInstance.end();
						return null;
					}
				}
				, iec, true);

	}

	/* (non-Javadoc)
	 * @see org.eclipse.ebpm.workflow.api.IControlAPI#restart(long)
	 */
	public void restart(final long processInstanceId, final ModifiedMessage[] modifiedMessages) {
		
		ProcessInstanceExecutionContext iec = ProcessEngine.buildProcessInstanceExecutionContext(processInstanceId);
		AtomicBoolean faulted = new AtomicBoolean();
		
		final long[] tokenIds = (long[])ProcessEngine.doWithinSpagic("ControlAPI.restart()", jbpmEngine, 
			new ISpagicJBPMCommand() {
			
			public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
					Session spagicSession, ProcessInstanceExecutionContext instanceExecutionContext) {
					
				long[] tokenIds = new long[modifiedMessages.length];
				try {
					// Retrieve all token ids of the modified messages
					for (int i = 0; i < modifiedMessages.length; i++) {
						
						TransitionInstance ti = (TransitionInstance)spagicSession.get(TransitionInstance.class, modifiedMessages[i].getIdJournal());
												
						// Extract the token id from the message id
						String[] tokens = ti.getIdMessage().split(":");
						if (tokens.length < 3)
							throw new SpagicWorkflowException("Unable to restart process instance with id: " + instanceExecutionContext.getSpagicInstanceId());
						if (!tokens[0].equals("JBPM") || !tokens[1].equals("TOK"))	
							throw new SpagicWorkflowException("Unable to restart process instance with id: " + instanceExecutionContext.getSpagicInstanceId());
						tokenIds[i] = Long.valueOf(tokens[2]);
						
					}
					
					ProcessInstance spagicProcessInstance = AuditDBUtils.getProcessInstanceById(spagicSession, instanceExecutionContext.getSpagicInstanceId());
					
					// Allow restarting only faulted processes
					if (spagicProcessInstance.getState() != ProcessInstanceStateConstants.PROCESS_FAULTED) {
						throw new SpagicWorkflowException("Unable to restart process instance with id: " + instanceExecutionContext.getSpagicInstanceId());
					}					
					
					spagicProcessInstance.setState(ProcessInstanceStateConstants.PROCESS_ACTIVE);
					spagicProcessInstance.setRestarted(true);
					spagicSession.save(spagicProcessInstance);					
				} catch (Throwable th) {
					if (th instanceof SpagicWorkflowException)
						throw (SpagicWorkflowException)th;
					else
						throw new SpagicWorkflowException("Unable to restart process instance with id: " + instanceExecutionContext.getSpagicInstanceId(), th);
				}
				return tokenIds;
			}
		}
		, iec, faulted);		
		
		
		if (!faulted.get())
			ProcessEngine.doWithinJBPM("ControlAPI.restart()", jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session spagicSession, ProcessInstanceExecutionContext instanceExecutionContext) {
						
						try {
							org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance = jbpmContext.getProcessInstanceForUpdate(instanceExecutionContext.getJbpmProcessInstanceId());

							// restart only automatic leaf tokens
							List<Token> automaticLeafTokens = new ArrayList<Token>();
							Token rootToken = jbpmProcessInstance.getRootToken();
							collectAutomaticLeafTokens(rootToken, automaticLeafTokens);
														
							for (Token token : automaticLeafTokens) {
								log.info("token " + token.getId() + " with end date " + token.getEnd() + " is undergoing restart.");
								
								// Before starting the execution, we must check if we need
								// to update the XML_MESSAGE variable
								for (int j = 0; j < tokenIds.length; j++) {
									if (tokenIds[j] == token.getId()) {
										jbpmProcessInstance.getContextInstance().setVariable(XML_MESSAGE, modifiedMessages[j].getXmlMessage(), token);
										log.info("token " + token.getId() + " with end date " + token.getEnd() + " is undergoing restart.");
										break;
									}
								}
								
								token.getNode().execute(new ExecutionContext(token));
								jbpmContext.save(token);
							}
						} catch (Throwable t) {
							log.error(" Error Restarting Process Instance [" + instanceExecutionContext.getSpagicInstanceId() + "]", t);
							throw new SpagicWorkflowException(" Error Restarting Process Instance [" + instanceExecutionContext.getSpagicInstanceId() + "]", t);
						}
						
						return null;
					}
				}
				, iec, true, faulted);
	}
	
	private void collectAutomaticLeafTokens(Token token, List<Token> automaticLeafTokens) {
		final Map childrenMap = token.getChildren();
		if (childrenMap.isEmpty()) {
			if(token.getNode().getAction() != null) {
				automaticLeafTokens.add(token);
			}
			return;
		}
		for (Object childObj : childrenMap.values()) {
			final Token child = (Token) childObj;
			collectAutomaticLeafTokens(child, automaticLeafTokens);
		}
	}
		
	public void stopOnAutomaticTask(long processInstanceId, boolean stop) {
		/*
		Session aSession = null;
		Transaction tx = null;

		try {
			aSession = ProcessEngine.getSession();
			tx = aSession.beginTransaction();

			ProcessInstance processInstance = AuditDBUtils.getProcessInstanceById(aSession, processInstanceId);
			processInstance.setProcessExecutionState(stop ? ProcessExecutionStateConstants.STOP_EXECUTION_INSTANCE : ProcessExecutionStateConstants.NORMAL_EXECUTION_INSTANCE);
			// ProcessExecutionStateCache
			// .getProcessExecutionState(
			// ProcessExecutionStateConstants.STOP_EXECUTION);
			aSession.save(processInstance);
			tx.commit();
		} catch (Throwable th) {
			if (tx != null) {
				tx.rollback();
			}
			throw new SpagicWorkflowException("Cannot Require Stop On Automatic Task", th);
		} finally {
			if (aSession != null && aSession.isOpen()) {
				aSession.close();
			}
		}
		*/

	}

	/**
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ebpm.workflow.api.IControlAPI#executeTask(long,
	 * java.lang.String, org.eclipse.ebpm.workflow.api.Variable[])
	 */
	public void executeTask(long processInstanceId, String taskName, Variable[] vars) {
		doSomethingOnTask(processInstanceId, taskName, vars, END_TASK);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ebpm.workflow.api.IControlAPI#executeTask(long, long,
	 * org.eclipse.ebpm.workflow.api.Variable[])
	 */
	public void executeTask(long processInstanceId, long taskId, Variable[] vars) {
		doSomethingOnTask(processInstanceId, taskId, vars, END_TASK);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ebpm.workflow.api.IControlAPI#executeTaskWithStatus(long,
	 * java.lang.String, java.lang.String, org.eclipse.ebpm.workflow.api.Variable[])
	 */
	public void executeTaskWithStatus(long processInstanceId, String taskName, String taskStatus, Variable[] vars) {
		org.eclipse.ebpm.workflow.api.ProcessInstance pi = jbpmEngine.getQueryAPI().getProcessInstanceById(processInstanceId);
		String nameOfTaskInstanceStatusVariable = ProcessEngine.getTaskInstanceStatusVariableName(pi.getName(), taskName);

		Variable[] toPassVariables = null;

		int varLenght = 0;
		if (vars != null)
			varLenght = vars.length;

		toPassVariables = new Variable[varLenght + 1];

		if (vars != null)
			System.arraycopy(vars, 0, toPassVariables, 0, varLenght);
		Variable statusVariable = new Variable();
		statusVariable.setName(nameOfTaskInstanceStatusVariable);
		statusVariable.setValue(taskStatus);

		toPassVariables[varLenght] = statusVariable;
		executeTask(processInstanceId, taskName, toPassVariables);

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ebpm.workflow.api.IControlAPI#executeTaskWithStatus(long, long,
	 * java.lang.String, org.eclipse.ebpm.workflow.api.Variable[])
	 */
	public void executeTaskWithStatus(long processInstanceId, long taskId, String taskStatus, Variable[] vars) {
		String taskName = ProcessEngine.getTaskNameByTaskId(taskId);
		executeTaskWithStatus(processInstanceId, taskName, taskStatus, vars);

	}

	/**
	 * Clean up pending task if instance finished.
	 * 
	 * @param jbpmProcessInstance
	 *          the jbpm process instance
	 */
	@SuppressWarnings("unchecked")
	public void cleanUpPendingTaskIfInstanceFinished(org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance) {
		if (jbpmProcessInstance.hasEnded()) {
			Collection<TaskInstance> taskInstances = (Collection<TaskInstance>) jbpmProcessInstance.getTaskMgmtInstance()
					.getTaskInstances();
			for (TaskInstance ti : taskInstances) {
				if (ti.getEnd() == null) {
					ti.cancel();
				}
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.eclipse.ebpm.workflow.api.IControlAPI#executeTaskByActor(java.lang.String,
	 * long, long, org.eclipse.ebpm.workflow.api.Variable[])
	 */
	public void executeTaskByActor(String actorId, long processInstanceId, long taskId, Variable[] vars) {
		if (actorId == null)
			throw new NullPointerException("actorId cannot be null!");
		doSomethingOnTaskByActor(actorId, processInstanceId, taskId, vars, END_TASK);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.eclipse.ebpm.workflow.api.IControlAPI#executeTaskByActor(java.lang.String,
	 * long, java.lang.String, org.eclipse.ebpm.workflow.api.Variable[])
	 */
	public void executeTaskByActor(String actorId, long processInstanceId, String taskName, Variable[] vars) {
		if (actorId == null)
			throw new NullPointerException("actorId cannot be null!");
		doSomethingOnTaskByActor(actorId, processInstanceId, taskName, vars, END_TASK);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ebpm.workflow.api.IControlAPI#getGlobalVariables(long)
	 */
	public Variable[] getGlobalVariables(final long processInstanceId) {
		
		ProcessInstanceExecutionContext iec = ProcessEngine.buildProcessInstanceExecutionContext(processInstanceId);
		
		return (Variable[]) ProcessEngine.doWithinJBPM("ControlAPI.getGlobalVariables()", jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session spagicSession, ProcessInstanceExecutionContext instanceExecutionContext) {
						
						Long jbpmProcessInstanceId = instanceExecutionContext.getJbpmProcessInstanceId();
						org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance = jbpmContext.loadProcessInstance(jbpmProcessInstanceId);
						
						@SuppressWarnings("unchecked")
						Map<String, Object> jbpmVariables = (Map<String, Object>) jbpmProcessInstance.getContextInstance().getVariables();

						List<Variable> returnVarList = new ArrayList<Variable>();
						if (jbpmVariables != null) {
							for (String key : jbpmVariables.keySet()) {
								String prefix = TASK_VARIABLE_NOTATION;
								if (key.indexOf(prefix) < 0) {
									Variable var = new Variable();
									var.setName(key);
									var.setValue(ProcessEngine.getVariableValue(key, jbpmVariables.get(key), jbpmProcessInstance));
									returnVarList.add(var);
								}
							}
						}
						Variable[] vars = returnVarList.toArray(new Variable[0]);
						return vars;
					}
				}, 
				iec,false);
	}
	

	
	
	
	
	
	
	public void setGlobalVariable(final long processInstanceId, final String variableName, final Object variableValue) {
		

		ProcessInstanceExecutionContext iec = ProcessEngine.buildProcessInstanceExecutionContext(processInstanceId);
		
		
		ProcessEngine.doWithinJBPM("ControlAPI.setGlobalVariable()", jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session spagicSession, ProcessInstanceExecutionContext instanceExecutionContext) {
							
						
						org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance = jbpmContext.loadProcessInstance(instanceExecutionContext.getJbpmProcessInstanceId());
						
						ProcessEngine.storeVariable(variableName, variableValue, jbpmProcessInstance);
						return null;
					}
				}, 
				iec,true);	
		
	}

	
	
	protected Long createProcessInstance(final String processName, Variable[] vars){
		
		return (Long)ProcessEngine.doWithinJBPMAndSpagic("createProcessInstance()", jbpmEngine, 
				new CreateProcessInstanceCommand(processName, vars), null, false);
	}

	
	protected Long startProcessInstance(final Long processInstanceId){
		
		ProcessInstanceExecutionContext iec = ProcessEngine.buildProcessInstanceExecutionContext(processInstanceId);
		
		ProcessEngine.doWithinJBPM("startProcessInstance()", jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session spagicSession, ProcessInstanceExecutionContext instanceExecutionContext) {
						
						org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance = jbpmContext.getProcessInstanceForUpdate(instanceExecutionContext.getJbpmProcessInstanceId());
						jbpmProcessInstance.signal();	
						return null;
					}
				}, 
				iec,true);
		
		return processInstanceId;
		
	}
	/*
	 * 
	 * @see
	 * org.eclipse.ebpm.workflow.api.IControlAPI#startByProcessName(java.lang.String)
	 */
	public Long startByProcessName(String processName) {
		return startByProcessName(processName, null);
	}

	public Long startByProcessName(String processName, Variable[] vars) {
		Long spagicProcessInstanceId = createProcessInstance(processName, vars);
		startProcessInstance(spagicProcessInstanceId);
		return spagicProcessInstanceId;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ebpm.workflow.api.IControlAPI#setGlobalVariables(long,
	 * org.eclipse.ebpm.workflow.api.Variable[])
	 */
	public void setGlobalVariables(long processInstanceId, Variable[] variables) {

		for (Variable variable : variables) {
			setGlobalVariable(processInstanceId, variable.getName(), variable.getValue());
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ebpm.workflow.api.IControlAPI#linkProcesses(long, long)
	 */
	public void linkProcesses(final long firstProcessInstanceId, final long secondProcessInstanceId) {
		
		ProcessEngine.doWithinSpagic("linkProcesses()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						ProcessInstance firstProcessInstance = AuditDBUtils.getProcessInstanceById(aSession, firstProcessInstanceId);
						ProcessInstance secondProcessInstance = AuditDBUtils.getProcessInstanceById(aSession, secondProcessInstanceId);

						if ((firstProcessInstance == null) || (secondProcessInstance == null)) {
							throw new SpagicWorkflowException(" Process instance with ID ["
									+ (firstProcessInstance == null ? firstProcessInstanceId : secondProcessInstanceId) + "] Not Found");
						}

						// Verify that the two instances are not already linked with different
						// processes
						if ((firstProcessInstance.getPoolInstance() != null) && (secondProcessInstance.getPoolInstance() != null)) {
							throw new SpagicWorkflowException(" Process instances with ID [" + firstProcessInstanceId + "] and ["
									+ secondProcessInstanceId + "] already linked");
						}

						if ((firstProcessInstance.getPoolInstance() == null) && (secondProcessInstance.getPoolInstance() == null)) {

							// Create the pool instance
							PoolInstance poolInstance = new PoolInstance();

							Set<ProcessInstance> processInstances = new LinkedHashSet<ProcessInstance>(2);
							processInstances.add(firstProcessInstance);
							processInstances.add(secondProcessInstance);
							poolInstance.setProcessInstances(processInstances);
							poolInstance.setPoolType(AuditDBUtils.getPoolTypeByName(aSession, ANONYMOUS_POOL_TYPE));

							String uniquePoolInstanceKey = computeUniquePoolInstanceKey("", firstProcessInstance);
							uniquePoolInstanceKey = computeUniquePoolInstanceKey(uniquePoolInstanceKey, secondProcessInstance);
							poolInstance.setUniquePoolInstanceKey(uniquePoolInstanceKey);
							aSession.save(poolInstance);

							// Associate the process instance to the iter instance
							firstProcessInstance.setPoolInstance(poolInstance);
							aSession.save(firstProcessInstance);

							secondProcessInstance.setPoolInstance(poolInstance);
							aSession.save(secondProcessInstance);

						} else {
							// Use an existing pool
							PoolInstance poolInstance = null;
							ProcessInstance processInstanceToLink = null;
							if (firstProcessInstance.getPoolInstance() != null) {
								poolInstance = firstProcessInstance.getPoolInstance();
								processInstanceToLink = secondProcessInstance;
							} else {
								poolInstance = secondProcessInstance.getPoolInstance();
								processInstanceToLink = firstProcessInstance;
							}

							Set<ProcessInstance> processInstances = poolInstance.getProcessInstances();
							processInstances.add(processInstanceToLink);

							String uniquePoolInstanceKey = poolInstance.getUniquePoolInstanceKey();
							uniquePoolInstanceKey = computeUniquePoolInstanceKey(uniquePoolInstanceKey, processInstanceToLink);
							poolInstance.setUniquePoolInstanceKey(uniquePoolInstanceKey);
							aSession.save(poolInstance);

							// Associate the process instance to the iter instance
							processInstanceToLink.setPoolInstance(poolInstance);
							aSession.save(processInstanceToLink);
						}
						
						return null;
					}
				}, 
				null);
	}

	/**
	 * Compute unique pool instance key.
	 * 
	 * @param existingKey
	 *          the existing key
	 * @param processInstanceToLink
	 *          the process instance to link
	 * 
	 * @return the string
	 */
	private String computeUniquePoolInstanceKey(String existingKey, ProcessInstance processInstanceToLink) {
		return existingKey + "[" + processInstanceToLink.getIdProcessInstance() + "]";
	}

	//
	// Internal Methods
	// JBPM - Related
	//

	/**
	 * Start jbpm task.
	 * 
	 * @param jbpmContext
	 *          the jbpm context
	 * @param jbpmProcessInstance
	 *          the jbpm process instance
	 * @param taskInstance
	 *          the task instance
	 * @param vars
	 *          the vars
	 */
	protected void startJbpmTaskInstanceByActor(String actorId, JbpmContext jbpmContext, org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance,
			TaskInstance taskInstance,  Variable[] vars) {

		taskInstance.setActorId(actorId);
		startJbpmTaskInstance(jbpmContext, jbpmProcessInstance, taskInstance, vars);

	}
	
	protected void startJbpmTaskInstance(JbpmContext jbpmContext, org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance,
			TaskInstance taskInstance,  Variable[] vars) {

		if (vars != null) {
			for (Variable var : vars) {
				ProcessEngine.storeVariable(var.getName(),var.getValue(), jbpmProcessInstance);
			}
		}
		
		taskInstance.start();
		jbpmContext.save(taskInstance);

	}

	/**
	 * Execute jbpm task.
	 * 
	 * @param jbpmContext
	 *          the jbpm context
	 * @param jbpmProcessInstance
	 *          the jbpm process instance
	 * @param taskInstance
	 *          the task instance
	 * @param vars
	 *          the vars
	 */
	protected void executeJbpmTask(JbpmContext jbpmContext, org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance,
			TaskInstance taskInstance, Variable[] vars) {

		if (vars != null) {
			for (Variable var : vars) {
				ProcessEngine.storeVariable(var.getName(), var.getValue(), jbpmProcessInstance);
			}
		}
		taskInstance.end();
		jbpmContext.save(taskInstance);

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ebpm.workflow.api.IControlAPI#getGlobalVariable(long,
	 * java.lang.String)
	 */
	public Variable getGlobalVariable(final long processInstanceId, final String variableName) {

		ProcessInstanceExecutionContext iec = ProcessEngine.buildProcessInstanceExecutionContext(processInstanceId);
		return(Variable)ProcessEngine.doWithinJBPM("ControlAPI.getGlobalVariable()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session spagicSession, ProcessInstanceExecutionContext instanceExecutionContext) {
					
						org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance = jbpmContext.getProcessInstance(instanceExecutionContext.getJbpmProcessInstanceId());
						
						Object o = jbpmProcessInstance.getContextInstance().getVariable(variableName);
						Variable var = null;
						if (o != null) {
							var = new Variable();
							var.setName(variableName);
							var.setValue(ProcessEngine.getVariableValue(variableName, o, jbpmProcessInstance));
						}
						
						return var;
					}
				}, 
				iec,false);
	}

	public Variable getTaskVariable(long processInstanceId, String taskName, String variableName) {
		JbpmContext jbpmContext = jbpmConfiguration.createJbpmContext();
		ProcessInstanceExecutionContext instanceExecutionContext = jbpmEngine.buildProcessInstanceExecutionContext(processInstanceId);
		org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance = jbpmContext.loadProcessInstance(instanceExecutionContext.getJbpmProcessInstanceId());
		TaskInstance jbpmTaskInstance = ProcessEngine.findTaskInstanceWithName(jbpmProcessInstance, taskName);
	
		checkProcessInstanceWithTaskInstance(jbpmContext, instanceExecutionContext,jbpmTaskInstance.getId() );
		String globalNameOfTaskVariable = TASK_VARIABLE_NOTATION + taskName + "." + variableName;
		return getGlobalVariable(processInstanceId, globalNameOfTaskVariable);
	}



			

	public Variable[] getTaskVariables(final long processInstanceId, final String taskName) {
		ProcessInstanceExecutionContext iec = ProcessEngine.buildProcessInstanceExecutionContext(processInstanceId);
		
		return (Variable[])ProcessEngine.doWithinJBPM("ControlAPI.getTaskVariables()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {

					public Object execute(ProcessEngine processEngine,
							JbpmContext jbpmContext, Session spagicSession, ProcessInstanceExecutionContext instanceExecutionContext) {
						
						
						org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance = jbpmContext.getProcessInstance(instanceExecutionContext.getJbpmProcessInstanceId());
						
						TaskInstance jbpmTaskInstance = ProcessEngine.findTaskInstanceWithName(jbpmProcessInstance, taskName);
						checkProcessInstanceWithTaskInstance(jbpmContext, instanceExecutionContext,jbpmTaskInstance.getId());
						@SuppressWarnings("unchecked")
						Map<String, Object> jbpmVariables = (Map<String, Object>) jbpmProcessInstance
								.getContextInstance().getVariables();
						List<Variable> returnVarList = new ArrayList<Variable>();
						if (jbpmVariables != null) {
							for (String key : jbpmVariables.keySet()) {
								String prefix = TASK_VARIABLE_NOTATION
										+ taskName + ".";
								if (key.indexOf(prefix) == 0) {
									String varName = key.substring(prefix
											.length());
									Variable var = new Variable();
									var.setName(varName);
									var.setValue(ProcessEngine
											.getVariableValue(varName,
													jbpmVariables.get(key),
													jbpmProcessInstance));
									returnVarList.add(var);
								}
							}
						}
						if (returnVarList.size() == 0)
							return new Variable[0];
						Variable[] vars = new Variable[returnVarList.size()];
						for (int i = 0; i < vars.length; i++) {
							vars[i] = returnVarList.get(i);
						}

						return vars;

					}
				},iec,false);
		}

	public void setTaskVariable(final long processInstanceId, final String taskName, final String variableName, final Object variableValue) {

		ProcessInstanceExecutionContext iec = ProcessEngine.buildProcessInstanceExecutionContext(processInstanceId);
		
		ProcessEngine.doWithinJBPM("ControlAPI.setTaskVariable()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					public Object execute(ProcessEngine processEngine,
							JbpmContext jbpmContext, Session spagicSession, ProcessInstanceExecutionContext instanceExecutionContext) {
				
						org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance = jbpmContext.loadProcessInstance(instanceExecutionContext.getJbpmProcessInstanceId());
						
						TaskInstance jbpmTaskInstance = ProcessEngine.findTaskInstanceWithName(jbpmProcessInstance, taskName);
						checkProcessInstanceWithTaskInstance(jbpmContext, instanceExecutionContext,jbpmTaskInstance.getId());
						
						String taskVariableName = TASK_VARIABLE_NOTATION + taskName + "." + variableName;
						
						ProcessEngine.storeVariable(taskVariableName, variableValue, jbpmProcessInstance);
						return null;
					}
			},iec,false);
	
	}

	public void setTaskVariables(long processInstanceId, String taskName, Variable[] variables) {
		for (Variable variable : variables) {
			setTaskVariable(processInstanceId, taskName, variable.getName(), variable.getValue());
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ebpm.workflow.api.IControlAPI#startTask(long, long,
	 * org.eclipse.ebpm.workflow.api.Variable[])
	 */
	public void startTask(long processInstanceId, long taskId, Variable[] vars) {
		doSomethingOnTask(processInstanceId, taskId, vars, START_TASK);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ebpm.workflow.api.IControlAPI#startTask(long, java.lang.String,
	 * org.eclipse.ebpm.workflow.api.Variable[])
	 */
	public void startTask(long processInstanceId, String taskName, Variable[] vars) {
		doSomethingOnTask(processInstanceId, taskName, vars, START_TASK);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ebpm.workflow.api.IControlAPI#startTaskByActor(java.lang.String,
	 * long, long, org.eclipse.ebpm.workflow.api.Variable[])
	 */
	public void startTaskByActor(String actorId, long processInstanceId, long taskId, Variable[] vars) {
		if (actorId == null)
			throw new NullPointerException("actorId cannot be null!");
		doSomethingOnTaskByActor(actorId, processInstanceId, taskId, vars, START_TASK);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ebpm.workflow.api.IControlAPI#startTaskByActor(java.lang.String,
	 * long, java.lang.String, org.eclipse.ebpm.workflow.api.Variable[])
	 */
	public void startTaskByActor(String actorId, long processInstanceId, String taskName, Variable[] vars) {
		doSomethingOnTaskByActor(actorId, processInstanceId, taskName, vars, START_TASK);
	}
	
	protected void completeJbpmTaskInstance(JbpmContext jbpmContext, org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance, TaskInstance jbpmTaskInstance, Variable[] vars){
		executeJbpmTask(jbpmContext, jbpmProcessInstance, jbpmTaskInstance, vars);
		cleanUpPendingTaskIfInstanceFinished(jbpmProcessInstance);
	}
	
	protected void completeJbpmTaskInstanceByActor(String actorId, JbpmContext jbpmContext, org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance, TaskInstance jbpmTaskInstance, Variable[] vars){
		String taskOwner = jbpmEngine.getTaskOwner(jbpmTaskInstance);
		
		if (taskOwner != null && !taskOwner.equalsIgnoreCase(actorId)){
				throw new SpagicWorkflowException("The Task with id ["+jbpmTaskInstance.getId()+"] is already owned by ["+taskOwner+"]");
		}else{
			jbpmTaskInstance.setActorId(actorId);
			completeJbpmTaskInstance(jbpmContext, jbpmProcessInstance, jbpmTaskInstance, vars);
		}
	}

	protected void doSomethingOnTask(final long processInstanceId, final String taskName, final Variable[] vars, final boolean endTask) {
		
		ProcessInstanceExecutionContext iec = ProcessEngine.buildProcessInstanceExecutionContext(processInstanceId);
		
		ProcessEngine.doWithinJBPM("ControlAPI.doSomethingOnTask()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					public Object execute(ProcessEngine processEngine,
							JbpmContext jbpmContext, Session spagicSession, ProcessInstanceExecutionContext instanceExecutionContext) {
		
							org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance = jbpmContext.loadProcessInstance(instanceExecutionContext.getJbpmProcessInstanceId());

							TaskInstance jbpmTaskInstance = ProcessEngine.findTaskInstanceWithName(jbpmProcessInstance, taskName);
							checkProcessInstanceWithTaskInstance(jbpmContext , instanceExecutionContext, jbpmTaskInstance.getId() );
							if (jbpmTaskInstance != null) {
								if (processEngine.isProfiled(jbpmTaskInstance))
										throw new SpagicWorkflowException("The task ["+taskName+"] is subject to security restriction use the API providing the actor Id");
								if (endTask) {
									completeJbpmTaskInstance(jbpmContext, jbpmProcessInstance, jbpmTaskInstance, vars);
								} else {
									startJbpmTaskInstance(jbpmContext, jbpmProcessInstance, jbpmTaskInstance,  vars);
								}
							} else {
								log.error("TaskInstance not found for task with name[" + taskName + "] and processInstanceId["
										+ processInstanceId + "]");
								throw new SpagicWorkflowException("TaskInstance not found for task with name[" + taskName
										+ "] and processInstanceId[" + processInstanceId + "]");
							}
							return null;
					}
		},iec,true);
	}

	protected void doSomethingOnTask(final long processInstanceId, final long taskId, final Variable[] vars, final boolean endTask) {
		
		ProcessInstanceExecutionContext iec = ProcessEngine.buildProcessInstanceExecutionContext(processInstanceId);
		ProcessEngine.doWithinJBPM("ControlAPI.doSomethingOnTask()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					public Object execute(ProcessEngine processEngine,
							JbpmContext jbpmContext, Session spagicSession, ProcessInstanceExecutionContext instanceExecutionContext) {
						
						org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance = jbpmContext.loadProcessInstance(instanceExecutionContext.getJbpmProcessInstanceId());

						TaskInstance jbpmTaskInstance = jbpmContext.getTaskInstance(taskId);
						checkProcessInstanceWithTaskInstance(jbpmContext, instanceExecutionContext, taskId);
						if (jbpmTaskInstance != null) {
							if (jbpmEngine.isProfiled(jbpmTaskInstance))
								throw new SpagicWorkflowException("The task ["+taskId+"] is subject to security restriction use the API providing the actor Id");
							if (endTask) {
								completeJbpmTaskInstance(jbpmContext, jbpmProcessInstance, jbpmTaskInstance, vars);
							} else {
								startJbpmTaskInstance(jbpmContext, jbpmProcessInstance, jbpmTaskInstance,  vars);
							}
						} else {
							log.error("TaskInstance not found for task with id[" + taskId + "] and processInstanceId["
									+ processInstanceId + "]");
							throw new SpagicWorkflowException("TaskInstance not found for task with id[" + taskId
									+ "] and processInstanceId[" + processInstanceId + "]");
						}
						return null;
					}
			},iec,true);
	}

	
	
	protected void doSomethingOnTaskByActor(final String actorId, final long processInstanceId, final long taskId, final Variable[] vars, final boolean endTask) {
		if (actorId == null)
			throw new NullPointerException("actorId cannot be null!");
		
		ProcessInstanceExecutionContext iec = ProcessEngine.buildProcessInstanceExecutionContext(processInstanceId);
		ProcessEngine.doWithinJBPM("ControlAPI.doSomethingOnTaskByActor()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					public Object execute(ProcessEngine processEngine,
							JbpmContext jbpmContext, Session spagicSession, ProcessInstanceExecutionContext instanceExecutionContext) {
						
						org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance = jbpmContext.loadProcessInstance(instanceExecutionContext.getJbpmProcessInstanceId());
						TaskInstance jbpmTaskInstance = jbpmContext.getTaskInstance(taskId);
						checkProcessInstanceWithTaskInstance(jbpmContext, instanceExecutionContext, taskId);
						if (jbpmTaskInstance != null) {
							jbpmEngine.checkUserOnTask(jbpmTaskInstance, actorId);
							if (endTask) {
								completeJbpmTaskInstanceByActor(actorId,jbpmContext, jbpmProcessInstance, jbpmTaskInstance,vars);
							} else {
								startJbpmTaskInstanceByActor(actorId, jbpmContext, jbpmProcessInstance, jbpmTaskInstance,  vars);
							}
						} else {
							log.error("TaskInstance not found for task with ID[" + taskId + "] and processInstanceId[" + processInstanceId+ "]");
							throw new SpagicWorkflowException("TaskInstance not found for task with ID[" + taskId + "] and processInstanceId["
									+ processInstanceId + "]");
						}
						return null;
					}
		},iec,true);
	}

	protected void doSomethingOnTaskByActor(final String actorId, final long processInstanceId, final String taskName, final Variable[] vars,
			final boolean endTask) {
		if (actorId == null)
			throw new NullPointerException("actorId cannot be null!");
		
		ProcessInstanceExecutionContext iec = ProcessEngine.buildProcessInstanceExecutionContext(processInstanceId);
		ProcessEngine.doWithinJBPM("ControlAPI.doSomethingOnTaskByActor()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					public Object execute(ProcessEngine processEngine,
							JbpmContext jbpmContext, Session spagicSession, ProcessInstanceExecutionContext instanceExecutionContext) {
						
						org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance = jbpmContext.loadProcessInstance(instanceExecutionContext.getJbpmProcessInstanceId());

						TaskInstance jbpmTaskInstance = ProcessEngine.findTaskInstanceWithName(jbpmProcessInstance, taskName);
						checkProcessInstanceWithTaskInstance(jbpmContext, instanceExecutionContext,jbpmTaskInstance.getId() );
			
						if (jbpmTaskInstance != null) {
							jbpmEngine.checkUserOnTask(jbpmTaskInstance, actorId);
							if (endTask) {
								completeJbpmTaskInstanceByActor(actorId,jbpmContext, jbpmProcessInstance, jbpmTaskInstance,vars);
							} else {
								startJbpmTaskInstanceByActor(actorId, jbpmContext, jbpmProcessInstance, jbpmTaskInstance,  vars);
							}
						} else {
							log.error("TaskInstance not found for task with ID[" + taskName + "] and processInstanceId[" + processInstanceId+ "]");
							throw new SpagicWorkflowException("TaskInstance not found for task with ID[" + taskName + "] and processInstanceId["
									+ processInstanceId + "]");
						}
						return null;
					}
		},iec,true);
		
	}

	/**
	 * @see org.eclipse.ebpm.workflow.api.IControlAPI#rejectTask(long, long)
	 */
	public void rejectTask(final long processInstanceId, final long taskId) {
		
		
		ProcessInstanceExecutionContext iec = ProcessEngine.buildProcessInstanceExecutionContext(processInstanceId);
		
		ProcessEngine.doWithinJBPM("ControlAPI.rejectTask()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					public Object execute(ProcessEngine processEngine,
							JbpmContext jbpmContext, Session spagicSession, ProcessInstanceExecutionContext instanceExecutionContext) {
						
						TaskInstance jbpmTaskInstance = jbpmContext.getTaskInstance(taskId);
						checkProcessInstanceWithTaskInstance(jbpmContext, instanceExecutionContext,taskId );
						
						if (jbpmTaskInstance != null) {
							
							if (jbpmTaskInstance.isOpen()) {
								String owner = jbpmEngine.getTaskOwner(jbpmTaskInstance);
								if ( owner != null )
									throw new SpagicWorkflowException("The Task ["+taskId+"] of process instance ["+processInstanceId+"] is owned by ["+owner+"] only this user could reject the task");
								jbpmTaskInstance.setActorId(null);
								jbpmTaskInstance.setStart(null);
							}
							jbpmContext.save(jbpmTaskInstance);
			
						} else {
							log.error("TaskInstance not found for task with ID[" + taskId + "] and processInstanceId[" + processInstanceId
									+ "]");
							throw new RuntimeException("TaskInstance not found for task with ID[" + taskId + "] and processInstanceId["
									+ processInstanceId + "]");
						}
						return null;					
					}
		},iec,true);
	
	}
	
	/**
	 * @see org.eclipse.ebpm.workflow.api.IControlAPI#rejectTask(long, long)
	 */
	public void rejectTaskByActor(final String actorId, final long processInstanceId, final long taskId) {
		
		ProcessInstanceExecutionContext iec = ProcessEngine.buildProcessInstanceExecutionContext(processInstanceId);
		
		ProcessEngine.doWithinJBPM("ControlAPI.rejectTaskByActor()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					public Object execute(ProcessEngine processEngine,
							JbpmContext jbpmContext, Session spagicSession, ProcessInstanceExecutionContext instanceExecutionContext) {
					TaskInstance jbpmTaskInstance = jbpmContext.getTaskInstance(taskId);
					checkProcessInstanceWithTaskInstance(jbpmContext, instanceExecutionContext,taskId );
					
					if (jbpmTaskInstance != null) {
						
						if (jbpmTaskInstance.isOpen()) {
							jbpmEngine.checkUserOnTask(jbpmTaskInstance, actorId);
							jbpmTaskInstance.setActorId(null);
							jbpmTaskInstance.setStart(null);
						}
						jbpmContext.save(jbpmTaskInstance);
		
					} else {
						log.error("TaskInstance not found for task with ID[" + taskId + "] and processInstanceId[" + processInstanceId
								+ "]");
						throw new RuntimeException("TaskInstance not found for task with ID[" + taskId + "] and processInstanceId["
								+ processInstanceId + "]");
					}
					return null;					
					}
		},iec,true);
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ebpm.workflow.api.IControlAPI#advanceTaskCompletion(long, long,
	 * int)
	 */
	@SuppressWarnings("unchecked")
	public void advanceTaskCompletion(final long processInstanceId, final long taskId, final int completionPercentage) {
		advanceTaskCompletionByActor(null, processInstanceId, taskId, completionPercentage);
	}

	
	@SuppressWarnings("unchecked")
	public void advanceTaskCompletion(final long processInstanceId, final String taskName, final int completionPercentage) {
		advanceTaskCompletionByActor(null,
											processInstanceId, 
											taskName, 
											completionPercentage);
		
	}
	
	@SuppressWarnings("unchecked")
	public void advanceTaskCompletionByActor(final String actorId, 
											 final long processInstanceId, 
											 final String taskName, 
											 final int completionPercentage) {
		
		ProcessInstanceExecutionContext iec = ProcessEngine.buildProcessInstanceExecutionContext(processInstanceId);
		
		Long taskInstanceId = (Long)
		ProcessEngine.doWithinJBPM("ControlAPI.advanceTaskCompletionByActor()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					public Object execute(ProcessEngine processEngine,
							JbpmContext jbpmContext, Session spagicSession, ProcessInstanceExecutionContext instanceExecutionContext) {
							
							org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance = jbpmContext.getProcessInstance(instanceExecutionContext.getJbpmProcessInstanceId());
							TaskInstance jbpmTaskInstance = ProcessEngine.findTaskInstanceWithName(jbpmProcessInstance, taskName);
							checkProcessInstanceWithTaskInstance(jbpmContext, instanceExecutionContext,jbpmTaskInstance.getId() );
							
							if (jbpmTaskInstance != null)
								return jbpmTaskInstance.getId();
							else
								return null;
					}
			},iec,false);
		if (taskInstanceId != null){
			advanceTaskCompletionByActor(actorId, processInstanceId, taskInstanceId, completionPercentage);
		}else{
			log.error("Task with name["+taskName+"] not found for Process ["+processInstanceId+"]");
		}
	
	}
	
	public void advanceTaskCompletionByActor(final String actorId,
											 final long processInstanceId, 
											 final long taskId, 
											 final int completionPercentage) {
		
		ProcessInstanceExecutionContext iec = ProcessEngine.buildProcessInstanceExecutionContext(processInstanceId);
		ProcessEngine.doWithinJBPM("ControlAPI.advanceTaskCompletionByActor()", 
				jbpmEngine, 
				new ISpagicJBPMCommand() {
					public Object execute(ProcessEngine processEngine,
							JbpmContext jbpmContext, Session spagicSession, ProcessInstanceExecutionContext instanceExecutionContext) {
		
						org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance = jbpmContext.getProcessInstance(instanceExecutionContext.getJbpmProcessInstanceId());
		
						TaskInstance jbpmTaskInstance = jbpmContext.getTaskInstance(taskId);
						checkProcessInstanceWithTaskInstance(jbpmContext, instanceExecutionContext,taskId);
						if (jbpmTaskInstance != null) {
							if (jbpmEngine.isProfiled(jbpmTaskInstance)){
								// The Task is profiled
								if (actorId == null)
									throw new SpagicWorkflowException("The task ["+taskId+"] is subject to security restriction use the API providing the actor Id");
								else
									jbpmEngine.checkUserOnTask(jbpmTaskInstance, actorId);
							}
							
							
							List<org.jbpm.graph.exe.Comment> comments = jbpmTaskInstance.getComments();
			
							
							// Retrieve the percentage comment and delete it
							for (Iterator<org.jbpm.graph.exe.Comment> it = comments.iterator(); it.hasNext();) {
								org.jbpm.graph.exe.Comment comment = it.next();
								if (comment.getMessage().startsWith(ProcessEngine.COMPLETION_PERCENTAGE_COMMENT)) {
									it.remove();
									break;
								}
							}
			
							// Add new percentage comment
							String comment = ProcessEngine.COMPLETION_PERCENTAGE_COMMENT + ProcessEngine.COMMENT_SEPARATION
									+ String.valueOf(completionPercentage);
							jbpmTaskInstance.addComment(comment);
			
							jbpmContext.save(jbpmTaskInstance);
						} else {
							log.error("TaskInstance not found for task with ID[" + taskId + "] and processInstanceId[" + processInstanceId
									+ "]");
							throw new RuntimeException("TaskInstance not found for task with ID[" + taskId + "] and processInstanceId["
									+ processInstanceId + "]");
						}
						return null;					
					}
				},iec,true);
		
	}
	
	protected void checkProcessInstanceWithTaskInstance(JbpmContext jbpmContext ,  ProcessInstanceExecutionContext instanceExecutionContext, Long taskInstanceId ){
		TaskInstance ti = jbpmContext.getTaskInstance(taskInstanceId);
		org.jbpm.graph.exe.ProcessInstance pi = ti.getProcessInstance();
		if (pi.getId() != instanceExecutionContext.getJbpmProcessInstanceId())
			throw new SpagicWorkflowException("The process instance is not associated with the task instance");
	}
	
	public void saveFDFtoDB(final FDFDocument fdf,final long processInstanceId,final String taskName){
			
		 ProcessEngine.doQueryWithinSpagic("QueryAPI.getProcessGraph()", jbpmEngine, new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session aSession,
							ProcessInstanceExecutionContext instanceExecutionContext) {
						
						ByteArrayOutputStream out = new ByteArrayOutputStream();
						try{
							fdf.save(out);
							AuditDBUtils.saveFDFtoDB(aSession, processInstanceId, taskName, out.toByteArray());
						}catch (Exception e) {
							log.error(e.getMessage(), e);
						}
						return new Object();
					}
				},null); 
			
	}
	
}
