/**

    Copyright 2007, 2010 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic.workflow.api.jbpm;

import java.util.Collection;
import java.util.List;
import java.util.Set;

import org.eclipse.ebpm.workflow.api.IControlExtendedAPI;
import org.eclipse.ebpm.workflow.api.SpagicWorkflowException;
import org.hibernate.Session;
import org.jbpm.JbpmConfiguration;
import org.jbpm.JbpmContext;
import org.jbpm.graph.def.Node;
import org.jbpm.graph.def.ProcessDefinition;
import org.jbpm.graph.def.Transition;
import org.jbpm.graph.exe.ExecutionContext;
import org.jbpm.graph.exe.Token;
import org.jbpm.graph.node.Fork;
import org.jbpm.graph.node.Join;
import org.jbpm.taskmgmt.exe.TaskMgmtInstance;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class ControlExtendedAPI implements IControlExtendedAPI {
	
	/** The Constant log. */
	private static Logger log = LoggerFactory.getLogger(ControlExtendedAPI.class);

	/** The jbpm engine. */
	private ProcessEngine jbpmEngine = null;
	
	/** The jbpm configuration. */
	private JbpmConfiguration jbpmConfiguration = null;

	
	/**
	 * Instantiates a new control extended api.
	 * 
	 * @param jbpmEngine
	 * @param jbpmConfiguration
	 */
	public ControlExtendedAPI(JbpmConfiguration jbpmConfiguration, ProcessEngine jbpmEngine) {
		super();
		this.jbpmEngine = jbpmEngine;
		this.jbpmConfiguration = jbpmConfiguration;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ebpm.workflow.api.IControlExtendedAPI#areNodesOnSameToken(long, java.lang.String, java.lang.String)
	 */
	public boolean areNodesOnSameToken(long processInstanceId, final String sourceTaskName, final String destTaskName) {
		
		ProcessInstanceExecutionContext iec = ProcessEngine.buildProcessInstanceExecutionContext(processInstanceId);
		return (Boolean)ProcessEngine.doWithinJBPM("ControlAPI.areNodesOnSameToken()", jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session spagicSession, ProcessInstanceExecutionContext instanceExecutionContext) {
						
						org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance = jbpmContext.getProcessInstance(instanceExecutionContext.getJbpmProcessInstanceId());
						return areNodesOnSameToken(jbpmProcessInstance.getProcessDefinition(), sourceTaskName, destTaskName);

					}
				}
				, iec, false);
	}

	protected boolean areNodesOnSameToken(ProcessDefinition pd, String sourceTaskName, String destTaskName) {
		
		Node sourceNode = pd.getNode(sourceTaskName);
		Node destNode = pd.getNode(destTaskName);
		
		log.debug("Checking compatibility between " + sourceTaskName + " and " + destTaskName);
		
		boolean result = false;
		try {
			Node lastNode = navigate(sourceNode, destNode, false);
			result = (destNode == lastNode);
		} catch (RuntimeException rt) {
			result = false;
		}
		
		return result;
	}
	
	/**
	 * Navigate within the process definition, from the sourceNode to the destNode, raising an
	 * exception if the nodes are not within the same token.
	 * @param sourceNode Start node
	 * @param destNode Destination node
	 * @param isWithinParallel Must be true if we are inside a parallel block
	 * @return The destination node, if reachable, otherwise the last visited node
	 */
	protected Node navigate(Node sourceNode, Node destNode, boolean isWithinParallel) {

		List<Transition> transitions = sourceNode.getLeavingTransitions();

		for (Transition tr : transitions) {
			Node nextNode = tr.getTo();
			
			// Follow entire current path until the DEST node is found
			Node lastNode = null;
			
			if (nextNode == destNode) {

				// Destination reached !!!
				if (isWithinParallel) {
					// We are within a parallel begun after the initial node:
					// not good !!
					throw new RuntimeException("Found the destination node within a parallel");
				} else {
					return destNode;
				}

			} else {
				// Advance until destination is reached, skipping fork nodes
				if (isForkNode(nextNode)) {
					// Skip the entire Fork-Join block
					lastNode = navigate(nextNode, destNode, true /* isWithinParallel */);
					
					// lastNode should be the Join node
					if (lastNode == null) {
						// Check next transition
						continue;
					} else if (isJoinNode(lastNode)) {
						// Wonderful !!!
						// We can continue the navigation
						lastNode = navigate(lastNode, destNode, isWithinParallel);
					} else {
						// Join node not found
						throw new RuntimeException("Found a Fork node without a correspondant Join");
					}
					
				} else if (isJoinNode(nextNode)) {
					if (isWithinParallel) {
						// The parallel is ended.
						// Proceed with forward search.
						return nextNode;
					} else {
						throw new RuntimeException("Found a Join node without a parallel");
					}
				} else {
					// Recursive call to manage automatically the multiple
					// destinations in the decision nodes
					lastNode = navigate(nextNode, destNode, isWithinParallel);
				}
			}
			
			if ((lastNode != null) && (lastNode == destNode)) {
				return destNode;				
			}
			
			// Check if we are waiting for a join node
			if (lastNode != null && isWithinParallel && isJoinNode(lastNode)) {
				return lastNode;
			}
		}
		// Destination unreachable
		return null;

	}
	
	
	protected boolean isForkNode(Node node) {
		return node instanceof Fork;
	}
	
	protected boolean isJoinNode(Node node) {
		return node instanceof Join;
	}
		
	/* (non-Javadoc)
	 * @see org.eclipse.ebpm.workflow.api.IControlExtendedAPI#forceToken(long, java.lang.String, java.lang.String)
	 */
	public void forceToken(final long processInstanceId, final String sourceTaskName, final String destTaskName) {
		
		ProcessInstanceExecutionContext iec = ProcessEngine.buildProcessInstanceExecutionContext(processInstanceId);
		
		ProcessEngine.doWithinJBPM("ControlAPI.forceToken()", jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session spagicSession, ProcessInstanceExecutionContext instanceExecutionContext) {
						
						org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance = jbpmContext.getProcessInstance(instanceExecutionContext.getJbpmProcessInstanceId());
						forceToken(jbpmContext, jbpmProcessInstance, sourceTaskName, destTaskName);
						jbpmContext.save(jbpmProcessInstance);
						return null;

					}
				}
				, iec, true);
		
		
		
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ebpm.workflow.api.IControlExtendedAPI#forceToken(long, java.lang.String)
	 */
	public void forceToken(final long processInstanceId, final String destTaskName) {
				
		ProcessInstanceExecutionContext iec = ProcessEngine.buildProcessInstanceExecutionContext(processInstanceId);
		
		ProcessEngine.doWithinJBPM("ControlAPI.forceToken()", jbpmEngine, 
				new ISpagicJBPMCommand() {
					
					public Object execute(ProcessEngine processEngine, JbpmContext jbpmContext,
							Session spagicSession, ProcessInstanceExecutionContext instanceExecutionContext) {
						
						org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance = jbpmContext.getProcessInstance(instanceExecutionContext.getJbpmProcessInstanceId());
						Token rootToken = jbpmProcessInstance.getRootToken();
			
						// Check that the root token hasn't active children
						if (rootToken.hasActiveChildren()) {
							// There is ambiguity on which is the startTaskName
							throw new SpagicWorkflowException("More than one active token in the process: you must specify the start token");
						}
			
						forceToken(jbpmContext, jbpmProcessInstance, rootToken.getNode().getName(), destTaskName);
						jbpmContext.save(jbpmProcessInstance);
						return null;
					}
			},iec,true);
	}
	
	private void forceToken(JbpmContext jbpmContext, org.jbpm.graph.exe.ProcessInstance jbpmProcessInstance, String sourceTaskName, String destTaskName) {
		
		Node destinationStep = null;
		Node sourceStep = null;
		
		if ((sourceTaskName == null) || (destTaskName == null)) {
			throw new SpagicWorkflowException("Start node name or destination node name cannot be null");
		}


			ProcessDefinition pd = jbpmProcessInstance.getProcessDefinition();
			
			// Determine if the step is a forward jump or a backward jump
			if (/* Forward jump */ areNodesOnSameToken(pd, sourceTaskName, destTaskName) ||
				/* Backward jump */	areNodesOnSameToken(pd, destTaskName, sourceTaskName)) {
				sourceStep = pd.getNode(sourceTaskName);
				destinationStep = pd.getNode(destTaskName);
			} else {
				// Jump not possible
				throw new SpagicWorkflowException("Start node and destination node are not within the same token");
			}
			
			// Determine the token of the start step
			List<Token> tokenList = jbpmProcessInstance.findAllTokens();
			Token tok = null;
			for (Token currentToken : tokenList) {
				if (currentToken.getNode().getId() == sourceStep.getId()) {
					tok = currentToken;
					break;
				}
			}
			if (tok == null) {
				throw new SpagicWorkflowException("Token for node:" + sourceTaskName + " not found");
			}
			
			TaskMgmtInstance tmi = jbpmProcessInstance.getTaskMgmtInstance();
			
			// End all unfinished tasks on current token
			Collection<org.jbpm.taskmgmt.exe.TaskInstance> unfinishedTasks = tmi.getUnfinishedTasks(tok);
			for (org.jbpm.taskmgmt.exe.TaskInstance ti : unfinishedTasks) {
				// Disable temporarily the signalling of the task instance, if enabled
				boolean isSignalling = ti.isSignalling();
				ti.setSignalling(false);
				ti.end();
				ti.setSignalling(isSignalling);
			}
									
			// Retrieve destination task and create a new task instance for it
			org.jbpm.taskmgmt.def.Task destinationTask = pd.getTaskMgmtDefinition().getTask(destinationStep.getName());


			// Determine a transition arriving to the destination node: in case of multiple
			// possible transitions, we select the first one. This is a force jump: this means
			// that there is no rule to determine the right transition in case of multiple arriving
			// transitions, because there is no right transition, we go directly to the destination
			// node, without a taken transition.
			Set<Transition> arrivingTransitions = destinationStep.getArrivingTransitions();
			if (arrivingTransitions.size() == 0) {
				// Should never happen
				throw new SpagicWorkflowException("No arriving transitions on node:" + destTaskName);
			}
			Transition firstTransition = arrivingTransitions.iterator().next();
			tok.setNode(firstTransition.getFrom());
			ExecutionContext exCo = new ExecutionContext(tok);
			exCo.setTransitionSource(firstTransition.getFrom());
			
			// Remove temporarily the transition from the condition to avoid errors
			// with conditioned transitions
			String transitionCondition = firstTransition.getCondition();
			firstTransition.setCondition(null);
			firstTransition.take(exCo);
			firstTransition.setCondition(transitionCondition);
		
	}


}
